<?php
namespace AIOSEO\Plugin\Addon\LinkAssistant\Main;

use AIOSEO\Plugin\Addon\LinkAssistant\Models;
use AIOSEO\Plugin\Common\Models as CommonModels;

// Exit if accessed directly.
if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

/**
 * Handles link extraction from posts.
 *
 * @since 1.0.0
 */
class Links {
	/**
	 * The action name of the links scan.
	 *
	 * @since 1.0.0
	 *
	 * @var string
	 */
	private $scanActionName = 'aioseo_link_assistant_links_scan';

	/**
	 * Class constructor.
	 *
	 * @since 1.0.0
	 */
	public function __construct() {
		add_action( $this->scanActionName, [ $this, 'scanPosts' ], 10, 1 );

		if ( ! is_admin() ) {
			return;
		}

		add_filter( 'save_post', [ $this, 'scanPost' ], 20, 1 );
		add_filter( 'init', [ $this, 'scheduleInitialScan' ], 3002 );
	}

	/**
	 * Schedules the initial links scan.
	 *
	 * @since 1.0.0
	 *
	 * @return void
	 */
	public function scheduleInitialScan() {
		aioseo()->helpers->scheduleSingleAction( $this->scanActionName, 10 );
	}

	/**
	 * Scans posts for links and stores them in the DB.
	 *
	 * @since 1.0.0
	 *
	 * @param  bool $shouldScheduleScan Whether a new scan should be scheduled.
	 * @return void
	 */
	public function scanPosts( $shouldScheduleScan = true ) {
		static $count = 0;
		$count++;

		aioseoLinkAssistant()->helpers->timeElapsed();

		$postsPerScan         = 10;
		$postTypes            = aioseoLinkAssistant()->helpers->getScannablePostTypes(); // Scan all post types so that results instantly show up when you include a new one.
		$implodedPostTypes    = aioseo()->helpers->implodeWhereIn( $postTypes, true );
		$postStatuses         = aioseo()->helpers->getPublicPostStatuses( true );
		$implodedPostStatuses = aioseo()->helpers->implodeWhereIn( $postStatuses, true );
		$minimumLinkScanDate  = aioseoLinkAssistant()->internalOptions->internal->minimumLinkScanDate;

		$posts = aioseo()->db->start( 'posts as p' )
			->select( 'DISTINCT p.ID, p.post_content, p.post_type, p.post_status' )
			->leftJoin( 'aioseo_posts as ap', 'p.ID = ap.post_id' )
			->whereRaw( "p.post_status IN ( $implodedPostStatuses )" )
			->whereRaw( "p.post_type IN ( $implodedPostTypes )" )
			->whereRaw( "(
				ap.post_id IS NULL OR
				ap.link_scan_date < p.post_modified_gmt OR
				ap.link_scan_date IS NULL OR
				ap.link_scan_date < '$minimumLinkScanDate'
			)" )
			->limit( $postsPerScan )
			->run()
			->result();

		if ( empty( $posts ) ) {
			aioseo()->helpers->scheduleSingleAction( $this->scanActionName, 15 * MINUTE_IN_SECONDS, [], true );

			return;
		}

		foreach ( $posts as $post ) {
			$this->scanPost( $post );
		}

		$timeElapsed = aioseoLinkAssistant()->helpers->timeElapsed();
		if ( 20 > $timeElapsed && 200 > $count ) {
			return $this->scanPosts( $shouldScheduleScan );
		}

		if ( $shouldScheduleScan ) {
			aioseo()->helpers->scheduleSingleAction( $this->scanActionName, 60, [], true );
		}
	}

	/**
	 * Scans the given post for links.
	 *
	 * @since 1.0.0
	 *
	 * @param  Object|int $post The post object or ID (if called on "save_post").
	 * @return void
	 */
	public function scanPost( $post ) {
		if ( ! is_object( $post ) ) {
			$post = aioseo()->helpers->getPost( $post );
		}

		if ( ! aioseoLinkAssistant()->helpers->isScannablePost( $post ) ) {
			return;
		}

		$this->update( $post->ID, $post->post_content );

		// Update the scan date in the posts table.
		$aioseoPost = CommonModels\Post::getPost( $post->ID );
		$aioseoPost->set( [
			'post_id'        => $post->ID,
			'link_scan_date' => gmdate( 'Y-m-d H:i:s' )
		] );
		$aioseoPost->save();
	}

	/**
	 * Updates Link records for the given post.
	 *
	 * @since 1.0.0
	 *
	 * @param  int    $postId      The post ID.
	 * @param  string $postContent The post content.
	 * @return void
	 */
	public function update( $postId, $postContent = null ) {
		if ( empty( $postContent ) ) {
			$post = aioseo()->helpers->getPost( $postId );
			if ( ! is_object( $post ) ) {
				return;
			}

			$postContent = $post->post_content;
		}

		$links = $this->extract( $postId, $postContent );

		// Delete all links first. We have to do this in order to remove old links that no longer exist.
		Models\Link::deleteLinks( $postId );

		if ( empty( $links ) ) {
			return;
		}

		$insertValues = [];
		foreach ( $links as $linkData ) {
			$data = Models\Link::sanitizeLink( $linkData );
			if ( empty( $data ) ) {
				continue;
			}

			if ( ! Models\Link::validateLink( $data ) ) {
				continue;
			}

			$insertValues[] = vsprintf(
				"(%d, %d, %d, %d, %d, '%s', '%s', '%s', '%s', '%s', '%s', '%s', NOW(), NOW())",
				$data
			);
		}

		$implodedInsertValues = implode( ',', $insertValues );

		$tableName = aioseo()->db->prefix . 'aioseo_links';
		aioseo()->db->execute(
			"INSERT INTO $tableName
			(`post_id`, `linked_post_id`, `internal`, `external`, `affiliate`, `url`, `hostname`, `anchor`, `phrase`, `phrase_html`, `paragraph`, `paragraph_html`, `created`, `updated`)
			VALUES $implodedInsertValues"
		);
	}

	/**
	 * Returns the existing links that are in the post content.
	 *
	 * @since 1.0.0
	 *
	 * @param  int    $postId      The post ID.
	 * @param  string $postContent The post content.
	 * @return array        The links.
	 */
	private function extract( $postId, $postContent ) {
		$postContent = aioseo()->helpers->pregReplace( '/&nbsp;/', ' ', $postContent );
		$postContent = aioseo()->helpers->decodeHtmlEntities( $postContent );

		/**
		 * Regex pattern divided into groups:
		 * 0  - Full phrase with link tag.
		 * 2  - Start of the phrase, before the anchor.
		 * 4  - The URL.
		 * 6  - The anchor.
		 * 9  - The end of the phrase, after the anchor.
		 * 10 - The ending punctuation mark.
		 */
		preg_match_all(
			'/(([^\r\n.?!]*)<t?a[^>]*?href=(\"|\')([^\"\']*?)(\"|\')[^>]*?>([\s\w\W]*?)<\/t?a>|<!-- wp:core-embed\/wordpress {"url":"([^"]*?)"[^}]*?"} -->|(?:>|&nbsp;|\s)((?:(?:http|ftp|https)\:\/\/)(?:[\w_-]+(?:(?:\.[\w_-]+)+))(?:[\w.,@?^=%&:\/~+#-]*[\w@?^=%&\/~+#-]))(?:<|&nbsp;|\s))([^<>.?!\r\n]*)([.?!]?)/i', // phpcs:disable Generic.Files.LineLength.MaxExceeded
			$postContent,
			$matches
		);

		if ( empty( $matches[0] ) ) {
			return [];
		}

		static $siteUrl = null;
		if ( null === $siteUrl ) {
			$siteUrl = wp_parse_url( get_site_url(), PHP_URL_HOST );
		}

		$links = [];
		foreach ( $matches[0] as $k => $v ) {
			if ( empty( $matches[4][ $k ] ) || empty( $matches[6][ $k ] ) ) {
				continue;
			}

			$url      = $matches[4][ $k ];
			$hostname = wp_parse_url( $url, PHP_URL_HOST );

			// If the URL is relative, add the hostname of the site.
			if ( ! $hostname ) {
				$hostname = $siteUrl;
				$url      = get_site_url() . $url;
			}

			// NOTE: We need to check this here before we strip off the "www" part.
			// Otherwise we will not be able to detect internal links on sites running on "www".
			$isInternal = $hostname === $siteUrl;

			$hostname = aioseo()->helpers->pregReplace( '/www\./i', '', $hostname );
			$anchor   = wp_strip_all_tags( $matches[6][ $k ] );
			// Remove trailing URL tags. The regex isn't sufficient for this.
			$phrase = wp_strip_all_tags( $matches[0][ $k ] );
			$phrase = trim( preg_replace( '/(.*)(<t?a[^<>].*$)/', '', $phrase ) );

			// Continue if the anchor or phrase are empty, e.g. empty link tag.
			if ( ! $anchor || ! $phrase ) {
				continue;
			}

			$phraseHtml = aioseo()->helpers->stripIncompleteHtmlTags( $matches[0][ $k ] );
			$phraseHtml = aioseo()->helpers->stripScriptTags( $phraseHtml );
			$phraseHtml = aioseoLinkAssistant()->helpers->trimParagraphTags( $phraseHtml );

			// For now, we'll drop list items as we can't parse them well.
			if ( empty( $phraseHtml ) || preg_match( '/<li|li>|<ol|ol>|<ul|ul>/', $phraseHtml ) ) {
				continue;
			}

			$paragraph     = aioseoLinkAssistant()->main->paragraph->get( $postId, $postContent, $phrase );
			$paragraphHtml = aioseoLinkAssistant()->main->paragraph->getHtml( $anchor, $paragraph, $postContent );
			$isAffiliate   = $this->isAffiliate( $url, $phraseHtml );

			$linkData = [
				'post_id'        => (int) $postId,
				'linked_post_id' => null,
				'internal'       => $isInternal,
				'external'       => ! $isInternal,
				'affiliate'      => $isAffiliate,
				'url'            => $url,
				'hostname'       => $hostname,
				'anchor'         => $anchor,
				'phrase'         => $phrase,
				'phrase_html'    => $phraseHtml,
				'paragraph'      => $paragraph,
				'paragraph_html' => $paragraphHtml
			];

			if ( $linkData['internal'] && ! $linkData['affiliate'] ) {
				$linkedPostId = url_to_postid( $linkData['url'] );
				if ( $linkedPostId ) {
					$linkData['linked_post_id'] = (int) $linkedPostId;
				}
			}

			$links[] = $linkData;
		}

		return $links;
	}

	/**
	 * Checks whether the given link is an affiliate link.
	 *
	 * @since 1.0.0
	 *
	 * @param  string $url        The URL.
	 * @param  string $phraseHtml The phrase HTML.
	 * @return bool               Whether the link is an affiliate link.
	 */
	private function isAffiliate( $url, $phraseHtml ) {
		static $affiliatePrefixes = null;
		if ( null === $affiliatePrefixes ) {
			$affiliatePrefixes = [];

			$prefixes = json_decode( aioseoLinkAssistant()->options->main->affiliatePrefix );
			if ( $prefixes ) {
				$affiliatePrefixes = array_map( function( $prefix ) {
					return $prefix->value;
				}, $prefixes );
			}
		}

		foreach ( $affiliatePrefixes as $affiliatePrefix ) {
			if ( aioseo()->helpers->stringContains( $url, $affiliatePrefix ) ) {
				return true;
			}
		}

		// Check if the link is wrapped in a ThirstyAffiliates tag.
		if ( preg_match( '/<ta[^>]*?href=(\"|\')([^\"\']*?)(\"|\')[^>]*?>[\s\w\W]*?<\/ta>/i', $phraseHtml ) ) {
			return true;
		}

		return false;
	}
}