<?php
namespace AIOSEO\Plugin\Addon\LinkAssistant\Main;

use AIOSEO\Plugin\Addon\LinkAssistant\Models;
use AIOSEO\Plugin\Common\Models as CommonModels;

// Exit if accessed directly.
if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

/**
 * The main class where all the scan magic happens.
 *
 * @since 1.0.0
 */
class Main {
	/**
	 * The action name of the posts scan.
	 *
	 * @since 1.0.0
	 *
	 * @var string
	 */
	private $scanActionName = 'aioseo_link_assistant_post_relationships_scan';

	/**
	 * Whether we should also insert inbound post relationship records.
	 * We have to do this when a post is created or updated since all posts that have already been scanned need to check against it.
	 *
	 * @since 1.0.0
	 *
	 * @var bool
	 */
	public $shouldInsertInboundPostRelationships = false;

	/**
	 * Class constructor.
	 *
	 * @since 1.0.0
	 */
	public function __construct() {
		if (
			! aioseo()->db->tableExists( 'aioseo_links' ) ||
			! aioseo()->db->tableExists( 'aioseo_links_suggestions' ) ||
			! aioseo()->db->tableExists( 'aioseo_links_post_relationships' )
		) {
			aioseolinkAssistant()->updates->addInitialTables();
			aioseolinkAssistant()->updates->addInitialColumns();

			return;
		}

		$this->links       = new Links();
		$this->paragraph   = new Paragraph();
		$this->suggestions = new Suggestions();

		$this->registerHooks();
	}

	/**
	 * Registers our hooks.
	 *
	 * @since 1.0.0
	 *
	 * @return void
	 */
	private function registerHooks() {
		add_action( 'save_post', [ $this, 'setShouldInsertInboundPostRelationships' ], 1, 0 );
		add_action( 'save_post', [ $this, 'scanPost' ], 20, 1 );

		add_action( 'delete_post', [ $this, 'removeOrphanedData' ] );

		add_action( $this->scanActionName, [ $this, 'scanPostRelationships' ], 10, 1 );

		if ( ! is_admin() ) {
			return;
		}

		add_filter( 'init', [ $this, 'scheduleInitialScan' ], 3002 );
	}

	/**
	 * Schedules the initial scan for suggestion scans.
	 *
	 * @since 1.0.0
	 *
	 * @return void
	 */
	public function scheduleInitialScan() {
		aioseo()->helpers->scheduleSingleAction( $this->scanActionName, 30 );
	}

	/**
	 * Checks which posts we need to schedule scans for.
	 *
	 * @since 1.0.0
	 *
	 * @return void
	 */
	public function scanPostRelationships() {
		// Scan all post types so that results instantly show up when you include new ones.
		$postTypes            = aioseoLinkAssistant()->helpers->getScannablePostTypes();
		$implodedPostTypes    = aioseo()->helpers->implodeWhereIn( $postTypes, true );
		$postStatuses         = aioseo()->helpers->getPublicPostStatuses( true );
		$implodedPostStatuses = aioseo()->helpers->implodeWhereIn( $postStatuses, true );

		$tableName    = aioseo()->db->db->prefix . 'aioseo_links_post_relationships';
		$subQuery     = "SELECT post_id from {$tableName} lss GROUP BY lss.post_id";
		$currentScans = aioseo()->db->execute( "SELECT count(*) as count FROM ($subQuery) AS scans", true )->result()[0]->count;

		$postsPerScan = $this->getPostsPerScan();
		// Store a limited amount of PostRelationships to keep the table small.
		if ( $postsPerScan * 10 <= $currentScans ) {
			return;
		}

		$minimumSuggestionScanDate = aioseoLinkAssistant()->internalOptions->internal->minimumSuggestionScanDate;

		$posts = aioseo()->db->start( 'posts as p' )
			->select( 'DISTINCT p.ID' )
			->leftJoin( 'aioseo_posts as ap', 'p.ID = ap.post_id' )
			->whereRaw( "p.post_status IN ( $implodedPostStatuses )" )
			->whereRaw( "p.post_type IN ( $implodedPostTypes )" )
			->whereRaw( "(
				ap.post_id IS NULL OR
				ap.link_suggestions_scan_date IS NULL OR
				ap.link_suggestions_scan_date < p.post_modified_gmt OR
				ap.link_suggestions_scan_date < '$minimumSuggestionScanDate'
			)" )
			->limit( $postsPerScan )
			->run()
			->result();

		if ( empty( $posts ) ) {
			$scanPercent = aioseoLinkAssistant()->helpers->getSuggestionsScanPercent();

			$scheduleActionTime = 100 === $scanPercent ? 15 * MINUTE_IN_SECONDS : 60;
			aioseo()->helpers->scheduleSingleAction( $this->scanActionName, $scheduleActionTime, [], true );

			return;
		}

		foreach ( $posts as $post ) {
			$this->scanPost( $post->ID, null );
		}

		aioseo()->helpers->scheduleSingleAction( $this->scanActionName, 60, [], true );

		$suggestionsScanActionName = 'aioseo_link_assistant_suggestions_scan';
		$actions = array_values( as_get_scheduled_actions( [
			'hook'     => $suggestionsScanActionName,
			'status'   => \ActionScheduler_Store::STATUS_PENDING,
			'per_page' => 1
		] ) );

		if ( ! empty( $actions ) ) {
			$schedule  = $actions[0]->get_schedule();
			$date      = $schedule->get_date();
			$timestamp = $date->getTimestamp();
			if ( ( time() - $timestamp ) > 90 ) {
				// If we just found results, we want to "wake up" the suggestions scan in case it's scheduled in more than a minute from now.
				as_unschedule_all_actions( $suggestionsScanActionName );
				as_enqueue_async_action( $suggestionsScanActionName, [], 'aioseo' );
			}
		}
	}

	/**
	 * Finds and stores post relationships for the given post.
	 *
	 * @since 1.0.0
	 *
	 * @param  int    $postId      The post ID.
	 * @param  string $postContent The post content.
	 * @return void
	 */
	public function scanPost( $postId, $postContent = null ) {
		$post = aioseo()->helpers->getPost( $postId );
		if ( ! aioseoLinkAssistant()->helpers->isScannablePost( $post ) ) {
			return;
		}

		$this->insertPostRelationships( $postId );

		$postContent = $postContent ? $postContent : $post->post_content;

		// Update the scan date and keywords in the posts table.
		$aioseoPost = CommonModels\Post::getPost( $postId );
		$aioseoPost->set( [
			'post_id'                    => $postId,
			'link_suggestions_scan_date' => gmdate( 'Y-m-d H:i:s' )
		] );
		$aioseoPost->save();
	}

	/**
	 * Checks whether the save_post hook is running.
	 * Because this runs with a lower priority, we can detect whether the hook is running before insertPostRelationships() runs on the other callback.
	 *
	 * @since 1.0.0
	 *
	 * @return void
	 */
	public function setShouldInsertInboundPostRelationships() {
		$this->shouldInsertInboundPostRelationships = true;
	}

	/**
	 * Stores the post relationshipsl in the database.
	 * This helps us offload the scan to spread out and have the most impact.
	 *
	 * @since 1.0.0
	 *
	 * @param  int  $postId The ID of the scanned post.
	 * @return void
	 */
	private function insertPostRelationships( $postId ) {
		$postId               = (int) $postId;
		$postTypes            = aioseoLinkAssistant()->helpers->getScannablePostTypes();
		$implodedPostTypes    = aioseo()->helpers->implodeWhereIn( $postTypes, true );
		$postStatuses         = aioseo()->helpers->getPublicPostStatuses( true );
		$implodedPostStatuses = aioseo()->helpers->implodeWhereIn( $postStatuses, true );

		$posts = aioseo()->db->start( 'posts' )
			->select( 'DISTINCT ID' )
			->whereRaw( "ID != $postId" )
			->whereRaw( "post_status IN ( $implodedPostStatuses )" )
			->whereRaw( "post_type IN ( $implodedPostTypes )" )
			->orderBy( 'ID DESC' )
			->run()
			->result();

		if ( empty( $posts ) ) {
			return;
		}

		$insertValues = [];
		foreach ( $posts as $postToCheckAgainst ) {
			$postToCheckAgainstId = (int) $postToCheckAgainst->ID;
			$insertValues[]       = "($postId, $postToCheckAgainstId, 1, NOW(), NOW() )";
		}

		if ( $this->shouldInsertInboundPostRelationships ) {
			foreach ( $posts as $postToScan ) {
				// If we're just updating a post, we need to rescan all other posts against this post to generate new inbound suggestions for it.
				$postToScanId   = (int) $postToScan->ID;
				$insertValues[] = "($postToScanId, $postId, 0, NOW(), NOW() )";
			}
		}

		$implodedInsertValues = implode( ',', $insertValues );

		$tableName = aioseo()->db->prefix . 'aioseo_links_post_relationships';
		aioseo()->db->execute(
			"INSERT INTO $tableName (`post_id`, `suggestion_post_id`, `prioritized`, `created`, `updated`)
			VALUES $implodedInsertValues
			ON DUPLICATE KEY UPDATE `post_id` = `post_id`;"
		);
	}

	/**
	 * If a post is deleted, we should also remove any Link Assistant data for it.
	 *
	 * @since 1.0.0
	 *
	 * @return void
	 */
	public function removeOrphanedData( $postId ) {
		Models\Link::deleteLinks( $postId );
		Models\PostRelationship::deletePostRelationships( $postId );
		Models\Suggestion::deleteSuggestions( $postId );
	}

	/**
	 * Returns the amount of posts we should process per scan.
	 *
	 * @since 1.0.0
	 *
	 * @return int The amount of posts we should process.
	 */
	private function getPostsPerScan() {
		if ( defined( 'AIOSEO_POSTS_PER_SCAN' ) && isset( AIOSEO_POSTS_PER_SCAN['postRelationships'] ) ) {
			return (int) AIOSEO_POSTS_PER_SCAN['postRelationships'];
		}

		$totalPosts = aioseoLinkAssistant()->helpers->getTotalScannablePosts();

		// Base on amount of posts because amount of scans we need to do grows exponentially.
		$postsPerScan = floor( 10000 / $totalPosts );
		$postsPerScan = $postsPerScan < 1 ? 1 : $postsPerScan;

		return $postsPerScan;
	}
}