<?php
namespace AIOSEO\Plugin\Addon\LinkAssistant\Main;

use AIOSEO\Plugin\Addon\LinkAssistant\Models;
use AIOSEO\Plugin\Common\Models as CommonModels;

// Exit if accessed directly.
if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

/**
 * Handles link suggestion discovery.
 *
 * @since 1.0.0
 */
class Suggestions {
	/**
	 * The base URL for the Link Suggestions microservice.
	 *
	 * @since 1.0.0
	 *
	 * @var string
	 */
	private $baseUrl = 'https://link-suggestions.aioseo.com/v1/';

	/**
	 * The cache key.
	 *
	 * @since 1.0.0
	 *
	 * @var string
	 */
	private $cacheKey = 'link_assistant_suggestions_posts_to_check';

	/**
	 * The post that is being scanned.
	 *
	 * @since 1.0.0
	 *
	 * @var Object
	 */
	private $scannedPost;

	/**
	 * The suggestions.
	 *
	 * @since 1.0.0
	 *
	 * @var array
	 */
	private $suggestions = [];

	/**
	 * The action name of the suggestions scan.
	 *
	 * @since 1.0.0
	 *
	 * @var string
	 */
	private $scanActionName = 'aioseo_link_assistant_suggestions_scan';

	/**
	 * Class constructor.
	 *
	 * @since 1.0.0
	 */
	public function __construct() {
		add_action( $this->scanActionName, [ $this, 'scanPosts' ], 10, 1 );

		if ( ! is_admin() ) {
			return;
		}

		add_action( 'init', [ $this, 'scheduleInitialScan' ] );
		add_action( 'save_post', [ $this, 'bustPostsCache' ] );
	}

	/**
	 * Schedules the initial suggestions scan.
	 *
	 * @since 1.0.0
	 *
	 * @return void
	 */
	public function scheduleInitialScan() {
		aioseo()->helpers->scheduleSingleAction( $this->scanActionName, 90 );
	}

	/**
	 * Busts the posts to check cache when a post is updated.
	 *
	 * @since 1.0.0
	 *
	 * @return void
	 */
	public function bustPostsCache() {
		aioseo()->cache->delete( $this->cacheKey );
	}

	/**
	 * Scan through the post relations in the suggestions scan table and look for potential link suggestions.
	 *
	 * @since 1.0.0
	 *
	 * @return void
	 */
	public function scanPosts() {
		static $count = 0;
		$count++;

		aioseoLinkAssistant()->helpers->timeElapsed();

		$scannedPost       = Models\PostRelationship::getNextPostToScan();
		$this->scannedPost = aioseo()->helpers->getPost( $scannedPost->post_id );

		// If there is no post, schedule the next scan based on whether we've completed scanning all content on the site.
		if ( ! $scannedPost->exists() ) {
			$scanPercent = aioseoLinkAssistant()->helpers->getSuggestionsScanPercent();

			$scheduleActionTime = 100 === $scanPercent ? 15 * MINUTE_IN_SECONDS : 60;
			aioseo()->helpers->scheduleSingleAction( $this->scanActionName, $scheduleActionTime, [], true );

			return;
		}

		$relationshipsPerScan = 5000;
		$postsToCheckAgainst  = Models\PostRelationship::getPostsToCheckAgainst( $scannedPost->post_id, $relationshipsPerScan );
		if ( empty( $postsToCheckAgainst ) ) {
			return;
		}

		$postsToCheckAgainstIds = array_map( function ( $postToCheckAgainst ) {
			return $postToCheckAgainst->suggestion_post_id;
		}, $postsToCheckAgainst );

		// Get the cached data for ALL posts only if (1) we're dealing with a prioritized post; and (2) the amount of posts we need to scan is lower than the total amount of scannable posts.
		$posts = $scannedPost->prioritized && ( aioseoLinkAssistant()->helpers->getTotalScannablePosts() > count( $postsToCheckAgainst ) )
			? $this->getCachedPosts()
			: $this->getPosts( $postsToCheckAgainstIds );

		// First, delete all non-dismissed suggestions (if any) so that we can get rid of outdated suggestions.
		Models\Suggestion::deleteNonDismissedSuggestions( $scannedPost->post_id, $postsToCheckAgainst );

		$suggestions = $this->getSuggestions( $posts );
		$this->parseSuggestions( $suggestions );

		// Finally, delete the PostRelationships we've just processed.
		$postRelationshipIds = array_map( function ( $postToCheckAgainst ) {
			return $postToCheckAgainst->id;
		}, $postsToCheckAgainst );

		Models\PostRelationship::deletePostRelationshipsByIds( $postRelationshipIds );

		// If we still have time, iterate.
		$timeElapsed = aioseoLinkAssistant()->helpers->timeElapsed();
		if ( 10 > $timeElapsed && 200 > $count ) {
			return $this->scanPosts();
		}

		if ( ! aioseoLinkAssistant()->main->shouldInsertInboundPostRelationships ) {
			// Don't schedule a new scan if we're just refreshing a post; otherwise we'll end up with two scans.
			aioseo()->helpers->scheduleSingleAction( $this->scanActionName, 60, [], true );
		}
	}

	/**
	 * Returns the cached data for the posts that we need to check against.
	 *
	 * @since 1.0.0
	 *
	 * @return array[Object] The post objects.
	 */
	private function getCachedPosts() {
		$posts = aioseo()->cache->get( $this->cacheKey );

		if ( empty( $posts ) ) {
			$query = $this->getBasePostsQuery();
			$posts = $query->run()
				->result();

			aioseo()->cache->update( $this->cacheKey, $posts );
		}

		return $posts;
	}

	/**
	 * Returns the data for the posts that we need to check against.
	 *
	 * @since 1.0.0
	 *
	 * @param  array[int]    $postsToCheckAgainstIds The IDs of the posts we need to check against.
	 * @return array[Object]                         The post objects.
	 */
	private function getPosts( $postsToCheckAgainstIds ) {
		$query = $this->getBasePostsQuery();
		$posts = $query->whereIn( 'ID', $postsToCheckAgainstIds )
			->run()
			->result();

		return $posts;
	}

	/**
	 * Returns the base posts query.
	 *
	 * @since 1.0.0
	 *
	 * @return Database The database instance.
	 */
	private function getBasePostsQuery() {
		$postTypes            = aioseoLinkAssistant()->helpers->getScannablePostTypes();
		$implodedPostTypes    = aioseo()->helpers->implodeWhereIn( $postTypes, true );
		$postStatuses         = aioseo()->helpers->getPublicPostStatuses( true );
		$implodedPostStatuses = aioseo()->helpers->implodeWhereIn( $postStatuses, true );

		return aioseo()->db->start( 'posts' )
			->select( 'ID, post_title, post_name' )
			->whereRaw( "post_type IN ( $implodedPostTypes )" )
			->whereRaw( "post_status IN ( $implodedPostStatuses )" )
			->where( 'ID !=', $this->scannedPost->ID );
	}

	/**
	 * Get the suggestions from the API.
	 *
	 * @since 1.0.0
	 *
	 * @param  array $posts The posts to find suggestions for.
	 * @return array        The suggestions.
	 */
	private function getSuggestions( $posts ) {
		$totalPosts = aioseoLinkAssistant()->helpers->getTotalScannablePosts();

		$requestBody = [
			'post'     => [
				'id'    => $this->scannedPost->ID,
				'title' => $this->scannedPost->post_title,
				'slug'  => $this->scannedPost->post_name,
			],
			'phrases'  => $this->getPhrases(),
			'titles'   => $this->getPostTitles( $posts ),
			'options'  => [
				'postTypes'       => aioseoLinkAssistant()->options->postTypes,
				'postStatuses'    => aioseoLinkAssistant()->options->postStatuses,
				'skipSentences'   => aioseoLinkAssistant()->options->skipSentences,
				'affiliatePrefix' => aioseoLinkAssistant()->options->affiliatePrefix,
				'excludePosts'    => aioseoLinkAssistant()->options->excludePosts,
				'ignoredWords'    => $this->getIgnoredWords()
			],
			'versions' => [
				'main'          => AIOSEO_VERSION,
				'linkAssistant' => AIOSEO_LINK_ASSISTANT_VERSION
			],
			'domain'   => aioseo()->helpers->getSiteDomain(),
			'totals'   => [
				'posts' => $totalPosts
			]
		];

		$response = wp_remote_post( $this->getUrl() . 'suggestions/', [
			'headers' => [
				'X-AIOSEO-License' => aioseo()->options->general->licenseKey,
				'Content-Type'     => 'application/json'
			],
			'body'    => wp_json_encode( $requestBody ),
			'timeout' => 60
		] );

		if ( 200 !== wp_remote_retrieve_response_code( $response ) ) {
			return [];
		}

		$responseBody = json_decode( wp_remote_retrieve_body( $response ) );

		return ! empty( $responseBody->suggestions ) ? $responseBody->suggestions : [];
	}

	/**
	 * Returns a list of post titles mapped to their IDs for the API.
	 *
	 * @since 1.0.0
	 *
	 * @param  array $posts The posts with their titles.
	 * @return array        Key-value list of post titles mapped to their IDs.
	 */
	private function getPostTitles( $posts ) {
		$titles = [];
		foreach ( $posts as $post ) {
			$titles[ $post->ID ] = $post->post_title;
		}

		return $titles;
	}

	/**
	 * Get a curated list of removed and added ignored words for the API to compare against.
	 * We only send the differences to limit the bandwidth we use.
	 *
	 * @since 1.0.0
	 *
	 * @return array The ignored word differences.
	 */
	private function getIgnoredWords() {
		$pattern              = '/([\.?!][\r\n\s]+|\r\n|\r|\n|\s{2,})/u';
		$wordsToIgnore        = array_map( 'trim', preg_split( $pattern, aioseoLinkAssistant()->options->main->wordsToIgnore, -1, PREG_SPLIT_NO_EMPTY ) );
		$defaultWordsToIgnore = array_map( 'trim', preg_split( $pattern, aioseoLinkAssistant()->options->main->getDefault( 'wordsToIgnore' ), -1, PREG_SPLIT_NO_EMPTY ) );

		$removed = array_diff( $defaultWordsToIgnore, $wordsToIgnore );
		$added   = array_diff( $wordsToIgnore, array_intersect( $defaultWordsToIgnore, $wordsToIgnore ) );

		return [
			'removed' => $removed,
			'added'   => $added
		];
	}

	/**
	 * Parses and stores the suggestions we've received from the API in the database.
	 *
	 * @since 1.0.0
	 *
	 * @param  array $apiSuggestions The suggestions from the API.
	 * @return void
	 */
	public function parseSuggestions( $apiSuggestions ) {
		if ( empty( $apiSuggestions ) ) {
			return;
		}

		$postContent = $this->getPostContent();

		$parsedSuggestions = [];
		foreach ( $apiSuggestions as $postId => $suggestions ) {
			if ( in_array( $postId, $this->getUsedPostIds(), true ) ) {
				continue;
			}

			foreach ( $suggestions as $suggestion ) {
				// It's possible we've matched with a tag attribute, e.g. the alt tag of an image. In that case, we can ignore it.
				$phrase = trim( wp_strip_all_tags( $suggestion->phrase ) );
				if ( empty( $phrase ) ) {
					continue;
				}

				$anchor = $this->getAnchor( $suggestion );
				if ( empty( $anchor ) ) {
					continue;
				}

				$phraseHtml = aioseoLinkAssistant()->helpers->trimParagraphTags( $suggestion->phrase );

				// Before we save the phrase, we must wrap the anchor around a link tag so that we can prefill it.
				static $permalink = [];
				if ( ! isset( $permalink[ $postId ] ) ) {
					$permalink[ $postId ] = get_permalink( $postId );
				}

				$url                = $permalink[ $postId ];
				$linkTag            = "<a href=\"${url}\">$1</a>";
				$escapedAnchor      = aioseo()->helpers->escapeRegex( $anchor );
				$phraseHtmlWithLink = preg_replace( "/(${escapedAnchor})/i", $linkTag, $phraseHtml, 1 );

				$parsedSuggestions[] = [
					'post_id'              => (int) $this->scannedPost->ID,
					'linked_post_id'       => (int) $postId,
					'anchor'               => $anchor,
					'phrase'               => $phrase,
					'phrase_html'          => $phraseHtmlWithLink,
					'original_phrase_html' => $phraseHtml,
					'paragraph'            => aioseoLinkAssistant()->main->paragraph->get( $this->scannedPost->ID, $postContent, $phrase ),
					'paragraph_html'       => aioseoLinkAssistant()->main->paragraph->getHtml( $anchor, $phrase, $postContent, true )
				];
			}
		}

		$insertValues = [];
		foreach ( $parsedSuggestions as $suggestionData ) {
			$data = Models\Suggestion::sanitizeSuggestion( $suggestionData );
			if ( empty( $data ) ) {
				continue;
			}

			if ( ! Models\Suggestion::validateSuggestion( $suggestionData ) ) {
				continue;
			}

			$insertValues[] = vsprintf(
				"(%d, %d, '%s', '%s', '%s', '%s', '%s', '%s', 0, NOW(), NOW())",
				$data
			);
		}

		$implodedInsertValues = implode( ',', $insertValues );
		if ( empty( $implodedInsertValues ) ) {
			return;
		}

		$tableName = aioseo()->db->prefix . 'aioseo_links_suggestions';
		aioseo()->db->execute(
			"INSERT INTO $tableName (`post_id`, `linked_post_id`, `anchor`, `phrase`, `phrase_html`, `original_phrase_html`, `paragraph`, `paragraph_html`, `dismissed`, `created`, `updated`)
			VALUES $implodedInsertValues"
		);
	}

	/**
	 * Returns the anchor for the current suggestion/phrase.
	 *
	 * @since 1.0.0
	 *
	 * @param  Object $suggestion The suggestion.
	 * @return string             The anchor.
	 */
	private function getAnchor( $suggestion ) {
		$phrase    = trim( wp_strip_all_tags( $suggestion->phrase ) );
		$firstWord = aioseo()->helpers->escapeRegex( $suggestion->words[0] );
		$lastWord  = aioseo()->helpers->escapeRegex( $suggestion->words[ count( $suggestion->words ) - 1 ] );

		preg_match( "/$firstWord.*$lastWord/u", $phrase, $anchor );
		if ( empty( $anchor[0] ) ) {
			return '';
		}

		$anchor = $anchor[0];

		// Prevent us from capturing partial words.
		$escapedAnchor = aioseo()->helpers->escapeRegex( $anchor );
		preg_match( "/${escapedAnchor}(\b|[^.\s\\r\\n!?-]*)/i", $phrase, $match );
		$fullAnchor = ! empty( $match[0] ) ? trim( $match[0] ) : $anchor;

		return $fullAnchor;
	}

	/**
	 * Extracts and returns the phrases from the current post.
	 *
	 * @since 1.0.0
	 *
	 * @return array The phrases.
	 */
	private function getPhrases() {
		$postContent = $this->getPostContent();

		$phrases = preg_split( '/([\.?!][\r\n\s]+|\r\n|\r|\n|\s{2,})/u', $postContent, -1, PREG_SPLIT_NO_EMPTY );
		$phrases = array_filter( $phrases, [ $this, 'isValidPhrase' ] );

		$skipSentences = aioseoLinkAssistant()->options->main->skipSentences;
		if ( $skipSentences ) {
			array_splice( $phrases, 0, absint( $skipSentences ) );
		}

		return $phrases;
	}

	/**
	 * Determines whether the given phrase is valid for our purposes.
	 *
	 * @since 1.0.0
	 *
	 * @param  string $phrase The phrase.
	 * @return bool           Whether the phrase is valid.
	 */
	private function isValidPhrase( $phrase ) {
		$phrase = trim( wp_strip_all_tags( $phrase ) );
		if ( 1 >= str_word_count( $phrase ) ) {
			return false;
		}

		// Ignore the phrase if it already contains a link or HTML tags that are to display.
		if ( preg_match( '/<a|a>|<h[0-9]|h[0-9]>|<li|li>|<ul|ul>|<ol|ol>/i', $phrase ) ) {
			return false;
		}

		// Ignore JSON.
		if ( preg_match( '/".*":\s".*"|".*": {/i', $phrase ) ) {
			return false;
		}

		return true;
	}

	/**
	 * Returns the post content after parsing it.
	 *
	 * @since 1.0.0
	 *
	 * @return string The post content.
	 */
	private function getPostContent() {
		static $postContent = [];
		if ( ! isset( $postContent[ $this->scannedPost->ID ] ) ) {
			$parsedContent = strip_shortcodes( $this->scannedPost->post_content );
			$parsedContent = aioseo()->helpers->pregReplace( '/&nbsp;/', ' ', $parsedContent );
			$parsedContent = aioseo()->helpers->decodeHtmlEntities( $parsedContent );

			$postContent[ $this->scannedPost->ID ] = $parsedContent;
		}

		return $postContent[ $this->scannedPost->ID ];
	}

	/**
	 * Returns the IDs of posts that are already linked to in the scanned post.
	 *
	 * @since 1.0.0
	 *
	 * @return array[int] The IDs.
	 */
	public function getUsedPostIds() {
		static $usedPostIds = [];
		if ( ! isset( $usedPostIds[ $this->scannedPost->ID ] ) ) {
			$links = Models\Link::getOutboundInternalLinks( $this->scannedPost->ID );

			$usedPostIds[ $this->scannedPost->ID ] = array_map( function ( $link ) {
				return $link->linked_post_id;
			}, $links );
		}

		return $usedPostIds[ $this->scannedPost->ID ];
	}

	/**
	 * Returns the URL for the Link Suggestions microservice.
	 *
	 * @since 1.0.0
	 *
	 * @return string The URL.
	 */
	private function getUrl() {
		if ( defined( 'AIOSEO_LINK_SUGGESTIONS_URL' ) ) {
			return AIOSEO_LINK_SUGGESTIONS_URL;
		}

		return $this->baseUrl;
	}
}