<?php
namespace AIOSEO\Plugin\Addon\LinkAssistant\Models;

use AIOSEO\Plugin\Common\Models as CommonModels;

// Exit if accessed directly.
if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

/**
 * The PostRelationship DB Model.
 *
 * @since 1.0.0
 */
class PostRelationship extends CommonModels\Model {
	/**
	 * The name of the table in the database, without the prefix.
	 *
	 * @since 1.0.0
	 *
	 * @var string
	 */
	protected $table = 'aioseo_links_post_relationships';

	/**
	 * Fields that should be numeric values.
	 *
	 * @since 1.0.0
	 *
	 * @var array
	 */
	protected $numericFields = [ 'id', 'post_id', 'suggestion_post_id' ];

	/**
	 * Fields that should be JSON encoded.
	 *
	 * @since 1.0.0
	 *
	 * @var array
	 */
	protected $jsonFields = [];

	/**
	 * Fields that should be boolean values.
	 *
	 * @since 1.0.0
	 *
	 * @var array
	 */
	protected $booleanFields = [ 'prioritized' ];


	/**
	 * Returns the ID of the next post that has to be scanned.
	 *
	 * @since 1.0.0
	 *
	 * @return int The post ID.
	 */
	public static function getNextPostToScan() {
		$refreshedPostId = aioseo()->cache->get( 'link_assistant_refreshed_post' );
		if ( ! empty( $refreshedPostId ) ) {
			$postRelationship          = new self();
			$postRelationship->id      = -1; // This needs to be set so that the instance passes the exists() check in the scan.
			$postRelationship->post_id = $refreshedPostId;

			aioseo()->cache->delete( 'link_assistant_refreshed_post' );

			return $postRelationship;
		}

		return aioseo()->db->start( 'aioseo_links_post_relationships' )
			->orderBy( 'prioritized DESC, RAND()' )
			->limit( 1 )
			->run()
			->model( 'AIOSEO\\Plugin\\Addon\\LinkAssistant\\Models\\PostRelationship' );
	}

	/**
	 * Returns the PostRelationship record with the given post ID.
	 *
	 * @since 1.0.0
	 *
	 * @param  int             $postId The post ID.
	 * @param  int             $limit  The limit on the number of posts. Default is 100.
	 * @return PostRelationship         The PostRelationship object.
	 */
	public static function getPostsToCheckAgainst( $postId, $limit = 100 ) {
		return aioseo()->db->start( 'aioseo_links_post_relationships' )
			->where( 'post_id', $postId )
			->limit( $limit, 0 )
			->run()
			->models( 'AIOSEO\\Plugin\\Addon\\LinkAssistant\\Models\\PostRelationship' );
	}

	/**
	 * Deletes the PostRelationship records with the given IDs.
	 *
	 * @since 1.0.0
	 *
	 * @param  array $ids The PostRelationship IDs.
	 * @return void
	 */
	public static function deletePostRelationshipsByIds( $ids ) {
		aioseo()->db->delete( 'aioseo_links_post_relationships' )
			->whereIn( 'id', $ids )
			->run();
	}

	/**
	 * Deletes all PostRelationship records for the post with the given ID.
	 *
	 * @since 1.0.0
	 *
	 * @param  int  $postId The post ID.
	 * @return void
	 */
	public static function deletePostRelationships( $postId ) {
		aioseo()->db->delete( 'aioseo_links_post_relationships' )
			->where( 'post_id', $postId )
			->run();
	}
}