<?php
namespace AIOSEO\Plugin\Addon\LinkAssistant\Utils;

// Exit if accessed directly.
if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

use AIOSEO\Plugin\Addon\LinkAssistant\Models;
use AIOSEO\Plugin\Addon\LinkAssistant\Traits as Traits;

/**
 * Contains helper functions.
 *
 * @since 1.0.0
 */
class Helpers {
	use Traits\Overview;
	use Traits\LinksReport;
	use Traits\DomainsReport;
	use Traits\PostSettings;
	use Traits\Debug;

	/**
	 * Gets the data for Vue.
	 *
	 * @since 1.0.0
	 *
	 * @param  string $page The current page.
	 * @return array        The data.
	 */
	public function getVueData( $data = [], $page = null ) {
		if ( 'link-assistant' === $page ) {
			return $this->getMenuData( $data );
		}
		if ( 'post' === $page ) {
			return $this->getPostData( $data );
		}

		return $data;
	}

	/**
	 * Adds the data for the settings menu.
	 *
	 * @since 1.0.0
	 *
	 * @param  array $data The data.
	 * @return array       The modified data.
	 */
	public function getMenuData( $data = [] ) {
		$data['linkAssistant'] = [
			'options'         => aioseoLinkAssistant()->options->all(),
			'internalOptions' => aioseoLinkAssistant()->internalOptions->all(),
			'overview'        => $this->getOverviewData(),
			'linksReport'     => $this->getLinksReportData(),
			'domainsReport'   => $this->getDomainsReportData(),
			'suggestionsScan' => [
				'percent' => aioseoLinkAssistant()->helpers->getSuggestionsScanPercent()
			]
		];

		return $data;
	}

	/**
	 * Returns the links/suggestion and their totals for the given post.
	 *
	 * @since 1.0.0
	 *
	 * @param  int    $postId The post ID.
	 * @param  int    $offset The offset.
	 * @param  int    $limit  The limit.
	 * @param  string $type   The type of link/suggestion.
	 * @return array          The data.
	 */
	public function getPostLinks( $postId, $offset = 0, $limit = 10, $type = '' ) {
		if ( $type ) {
			return $this->getPostLinksHelper( $postId, $type, $offset, $limit );
		}

		$types = [
			'inboundInternal',
			'outboundInternal',
			'affiliate',
			'external',
			'suggestionsInbound',
			'suggestionsOutbound'
		];

		$data = [];
		foreach ( $types as $type ) {
			$data[ $type ] = $this->getPostLinksHelper( $postId, $type, $offset, $limit );
		}

		return $data;
	}

	/**
	 * Returns links/suggestions and their totals of the given type for the given post.
	 *
	 * @since 1.0.0
	 *
	 * @param  string $type   The link/suggestion type.
	 * @param  int    $offset The offset.
	 * @param  int    $limit  The limit.
	 * @return array          The data.
	 */
	private function getPostLinksHelper( $postId, $type, $offset = 0, $limit = 10 ) {
		switch ( $type ) {
			case 'inboundInternal':
				$total = Models\Link::getTotalInboundInternalLinks( $postId );

				return [
					'rows'   => array_values( Models\Link::getInboundInternalLinks( $postId, $offset, $limit ) ),
					'totals' => [
						'page'  => 0 === $offset ? 1 : ( $offset / 10 ) + 1,
						'pages' => ceil( $total / 10 ),
						'total' => $total
					]
				];
			case 'outboundInternal':
				$total = Models\Link::getTotalOutboundInternalLinks( $postId );

				return [
					'rows'   => array_values( Models\Link::getOutboundInternalLinks( $postId, $offset, $limit ) ),
					'totals' => [
						'page'  => 0 === $offset ? 1 : ( $offset / 10 ) + 1,
						'pages' => ceil( $total / 10 ),
						'total' => $total
					]
				];
			case 'affiliate':
				$total = Models\Link::getTotalAffiliateLinks( $postId );

				return [
					'rows'   => array_values( Models\Link::getAffiliateLinks( $postId, $offset, $limit ) ),
					'totals' => [
						'page'  => 0 === $offset ? 1 : ( $offset / 10 ) + 1,
						'pages' => ceil( $total / 10 ),
						'total' => $total
					]
				];
			case 'external':
				$total = Models\Link::getTotalExternalLinks( $postId );

				return [
					'rows'   => array_values( Models\Link::getExternalLinks( $postId, $offset, $limit ) ),
					'totals' => [
						'page'  => 0 === $offset ? 1 : ( $offset / 10 ) + 1,
						'pages' => ceil( $total / 10 ),
						'total' => $total
					]
				];
			case 'suggestionsOutbound':
				$total = Models\Suggestion::getTotalOutboundSuggestions( $postId );

				return [
					'rows'   => array_values( Models\Suggestion::getOutboundSuggestions( $postId, $offset, $limit ) ),
					'totals' => [
						'page'  => 0 === $offset ? 1 : ( $offset / 10 ) + 1,
						'pages' => ceil( $total / 10 ),
						'total' => $total
					]
				];
			case 'suggestionsInbound':
				$total = Models\Suggestion::getTotalInboundSuggestions( $postId );

				return [
					'rows'   => array_values( Models\Suggestion::getInboundSuggestions( $postId, $offset, $limit ) ),
					'totals' => [
						'page'  => 0 === $offset ? 1 : ( $offset / 10 ) + 1,
						'pages' => ceil( $total / 10 ),
						'total' => $total
					]
				];
			default:
				return [];
		}
	}

	/**
	 * Checks if the given post is excluded from Link Assistant.
	 *
	 * @since 1.0.0
	 *
	 * @param  int  $postId The post ID.
	 * @return bool         Whether the post is excluded.
	 */
	public function isExcludedPost( $postId ) {
		$excludedPostIds      = $this->getExcludedPostIds();
		$includedPostTypes    = $this->getIncludedPostTypes();
		$includedPostStatuses = $this->getIncludedPostStatuses();
		$post                 = aioseo()->helpers->getPost( $postId );

		return in_array( (int) $postId, $excludedPostIds, true ) ||
			! in_array( $post->post_type, $includedPostTypes, true ) ||
			! in_array( $post->post_status, $includedPostStatuses, true );
	}

	/**
	 * Returns the IDs of posts that are excluded from Link Assistant.
	 *
	 * @since 1.0.0
	 *
	 * @return array The post IDs.
	 */
	public function getExcludedPostIds() {
		static $excludedPostIds = null;
		if ( null === $excludedPostIds ) {
			$excludedPostIds = [];
			$excludedPosts   = aioseoLinkAssistant()->options->main->excludePosts;
			foreach ( $excludedPosts as $excludedPost ) {
				$excludedPost = json_decode( $excludedPost );
				if ( ! empty( $excludedPost->value ) ) {
					$excludedPostIds[] = $excludedPost->value;
				}
			}
		}

		return $excludedPostIds;
	}

	/**
	 * Returns the post types that Link Assistant is enabled for.
	 *
	 * @since 1.0.0
	 *
	 * @return array The included post types.
	 */
	public function getIncludedPostTypes() {
		static $includedPostTypes = null;

		if ( null !== $includedPostTypes ) {
			return $includedPostTypes;
		}

		$includedPostTypes = [];
		$postTypes         = aioseoLinkAssistant()->options->main->postTypes->all();
		if ( ! empty( $postTypes['all'] ) ) {
			$includedPostTypes = $this->getScannablePostTypes();
		} else {
			// Determine the intersection to make sure that we only consider post types that are currently registered.
			$includedPostTypes = array_intersect(
				$postTypes['included'],
				$this->getScannablePostTypes()
			);
		}

		foreach ( $includedPostTypes as $k => $postType ) {
			$capabilities = $this->getPostTypeCapabilities( $postType );
			if ( ! current_user_can( $capabilities['edit_posts'] ) ) {
				unset( $includedPostTypes[ $k ] );
			}
		}

		$includedPostTypes = apply_filters( 'aioseo_link_assistant_post_types', $includedPostTypes );

		return $includedPostTypes;
	}

	/**
	 * Returns a list of capabilities for the given post type.
	 *
	 * @since 1.0.2
	 *
	 * @param  string $postType The name of the post type.
	 * @return array            The capabilities.
	 */
	private function getPostTypeCapabilities( $postType ) {
		$postTypeObject = get_post_type_object( $postType );

		if ( ! is_array( $postTypeObject->capability_type ) ) {
			$postTypeObject->capability_type = [
				$postTypeObject->capability_type,
				$postTypeObject->capability_type . 's'
			];
		}

		// Singular base for meta capabilities, plural base for primitive capabilities.
		list( $singularBase, $pluralBase ) = $postTypeObject->capability_type;

		$capabilities = [
			'edit_post'          => 'edit_' . $singularBase,
			'read_post'          => 'read_' . $singularBase,
			'delete_post'        => 'delete_' . $singularBase,
			'edit_posts'         => 'edit_' . $pluralBase,
			'edit_others_posts'  => 'edit_others_' . $pluralBase,
			'delete_posts'       => 'delete_' . $pluralBase,
			'publish_posts'      => 'publish_' . $pluralBase,
			'read_private_posts' => 'read_private_' . $pluralBase,
		];

		return $capabilities;
	}


	/**
	 * Returns the post statuses that Link Assistant is enabled for.
	 *
	 * @since 1.0.0
	 *
	 * @return array The included post statuses.
	 */
	public function getIncludedPostStatuses() {
		static $includedPostStatuses = null;

		if ( null !== $includedPostStatuses ) {
			return $includedPostStatuses;
		}

		$includedPostStatuses = [];
		$postStatuses         = aioseoLinkAssistant()->options->main->postStatuses->all();
		if ( ! empty( $postStatuses['all'] ) ) {
			$includedPostStatuses = aioseo()->helpers->getPublicPostStatuses( true );
		} else {
			// Determine the intersection to make sure that we only consider post statuses that are currently registered.
			$includedPostStatuses = array_intersect(
				$postStatuses['included'],
				aioseo()->helpers->getPublicPostStatuses( true )
			);
		}

		$includedPostStatuses = apply_filters( 'aioseo_link_assistant_post_statuses', $includedPostStatuses );

		return $includedPostStatuses;
	}

	/**
	 * Deletes one or multiple links in the post content of a given post.
	 *
	 * @since 1.0.0
	 *
	 * @param  array|int $linkIds The Link IDs.
	 * @return void
	 */
	public function deleteLinksInPost( $linkIds ) {
		// NOTE: We can't save the post content until all links are deleted as it will trigger a re-scan on "save_post".
		if ( ! is_array( $linkIds ) ) {
			$linkIds = [ $linkIds ];
		}

		// We must reset the keys first since the WpTable component uses the model ID as the key for each record.
		$linkIds = array_values( $linkIds );
		if ( empty( $linkIds ) ) {
			return false;
		}

		$posts = [];
		foreach ( $linkIds as $linkId ) {
			$link = Models\Link::getLinkById( $linkId );
			if ( ! $link->exists() ) {
				continue;
			}

			if ( ! isset( $posts[ $link->post_id ] ) ) {
				$post = get_post( $link->post_id );

				// Replace encoded spaces with non-encoded ones first so that the paragraph can be found.
				$posts[ $link->post_id ] = preg_replace( '/&nbsp;/', ' ', $post->post_content );
			}

			$postContent = $posts[ $link->post_id ];

			// Strip the HTML link tag from the anchor in the paragraph.
			$escapedAnchor        = aioseo()->helpers->escapeRegex( $link->anchor );
			$pattern              = "/(<a[^<>]*>)[\\r\\n\s]*($escapedAnchor)[\\r\\n\\s]*(<\/a[^<>]*>)/i";
			$paragraphWithoutLink = preg_replace( $pattern, '$2', $link->paragraph_html );
			if ( preg_match( $pattern, $paragraphWithoutLink ) ) {
				continue;
			}

			// Replace the paragraph in the post content with the new one.
			$escapedParagraph        = aioseo()->helpers->escapeRegex( $link->paragraph_html );
			$posts[ $link->post_id ] = preg_replace( "/$escapedParagraph/i", $paragraphWithoutLink, $postContent );

			Models\Link::deleteLinkById( $linkId );
		}

		foreach ( $posts as $postId => $postContent ) {
			// Finally, we update the post with the modified post content.
			wp_update_post( [
				'ID'           => $postId,
				'post_content' => $postContent
			], true );
		}
	}

	/**
	 * Trims HTML paragraph from the start/end of the given string.
	 *
	 * @since 1.0.0
	 *
	 * @param  string $string The string.
	 * @return string         The modified string.
	 */
	public function trimParagraphTags( $string ) {
		$string = preg_replace( '/^<p[^>]*>/', '', $string );
		$string = preg_replace( '/<\/p>/', '', $string );

		return trim( $string );
	}

	/**
	 * Get the percent completed for the suggestions scan.
	 *
	 * @since 1.0.0
	 *
	 * @return int The percent completed as an integer.
	 */
	public function getSuggestionsScanPercent() {
		$postTypes                 = $this->getScannablePostTypes(); // Scan all post types so that results instantly show up when you include new ones.
		$implodedPostTypes         = aioseo()->helpers->implodeWhereIn( $postTypes, true );
		$postStatuses              = aioseo()->helpers->getPublicPostStatuses( true );
		$implodedPostStatuses      = aioseo()->helpers->implodeWhereIn( $postStatuses, true );
		$minimumSuggestionScanDate = aioseoLinkAssistant()->internalOptions->internal->minimumSuggestionScanDate;

		// If this isn't set yet, the plugin has just been installed and the full site still needs to be scanned.
		// The option should be set on the next request.
		if ( empty( $minimumSuggestionScanDate ) ) {
			return 0;
		}

		$tableName          = aioseo()->db->prefix . 'aioseo_links_post_relationships';
		$remainingScanCount = aioseo()->db->start( 'posts as p' )
			->leftJoin( 'aioseo_posts as ap', 'p.ID = ap.post_id' )
			->leftJoin( "(SELECT post_id FROM $tableName WHERE prioritized = 1 GROUP BY post_id) as alss", 'p.ID = alss.post_id', true )
			->whereRaw( "p.post_status IN ( $implodedPostStatuses )" )
			->whereRaw( "p.post_type IN ( $implodedPostTypes )" )
			->whereRaw( "(
				ap.post_id IS NULL OR
				ap.link_suggestions_scan_date IS NULL OR
				ap.link_suggestions_scan_date < p.post_modified_gmt OR
				ap.link_suggestions_scan_date < '$minimumSuggestionScanDate' OR
				alss.post_id IS NOT NULL
			)" )
			->groupBy( 'p.ID' )
			->count();

		$allPostsCount = aioseo()->db->start( 'posts as p' )
			->whereRaw( "p.post_status IN ( $implodedPostStatuses )" )
			->whereRaw( "p.post_type IN ( $implodedPostTypes )" )
			->count();

		if ( 0 === $allPostsCount || 0 === $remainingScanCount ) {
			return 100;
		}

		return round( 100 - 100 * ( $remainingScanCount / $allPostsCount ) );
	}

	/**
	 * Returns the time that elapsed since the initial call to this function.
	 *
	 * @since 1.0.0
	 *
	 * @return int|null The time that has elapsed.
	 */
	public function timeElapsed() {
		static $last = null;

		$now    = microtime( true );
		$return = null !== $last ? $now - $last : null;

		if ( null === $last ) {
			$last = $now;
		}

		return $return;
	}

	/**
	 * Returns the post title or a placeholder if there isn't one.
	 *
	 * @since 1.0.0
	 *
	 * @param  int    $postId The post ID.
	 * @return string         The post title.
	 */
	public function getPostTitle( $postId ) {
		static $titles = [];
		if ( isset( $titles[ $postId ] ) ) {
			return $titles[ $postId ];
		}

		$post  = aioseo()->helpers->getPost( $postId );
		$title = $post->post_title;
		$title = $title ? $title : __( '(no title)', 'default' );

		$titles[ $postId ] = aioseo()->helpers->decodeHtmlEntities( $title );

		return $titles[ $postId ];
	}

	/**
	 * Returns the total amount of posts on the site that are scannable.
	 * NOTE: This will always return a minimum of 1 to prevent division by 0 PHP errors.
	 *
	 * @since 1.0.0
	 *
	 * @return int The total amount of scannable posts.
	 */
	public function getTotalScannablePosts() {
		$totalPosts = aioseo()->cache->get( 'link_assistant_suggestions_scan_total_posts' );
		if ( null === $totalPosts ) {
			$postTypes            = $this->getScannablePostTypes();
			$implodedPostTypes    = aioseo()->helpers->implodeWhereIn( $postTypes, true );
			$postStatuses         = aioseo()->helpers->getPublicPostStatuses( true );
			$implodedPostStatuses = aioseo()->helpers->implodeWhereIn( $postStatuses, true );

			$totalPosts = aioseo()->db->start( 'posts as p' )
				->whereRaw( "p.post_status IN ( $implodedPostStatuses )" )
				->whereRaw( "p.post_type IN ( $implodedPostTypes )" )
				->count();

			if ( empty( $totalPosts ) ) {
				$totalPosts = 1;
			}

			aioseo()->cache->update( 'link_assistant_suggestions_scan_total_posts', $totalPosts );
		}

		return $totalPosts;
	}

	/**
	 * Applies wp_kses_post on the given string, but also allows some other tags we support.
	 *
	 * @since 1.0.1
	 *
	 * @param  string $string The string.
	 * @return string         The sanitized string.
	 */
	public function wpKsesPhrase( $string ) {
		$allowedHtmlTags = wp_kses_allowed_html( 'post' );

		$customTags = [
			'ta' => [
				'linkid' => [],
				'href'   => []
			]
		];

		$allowedHtmlTags = array_merge( $allowedHtmlTags, $customTags );

		return wp_kses( $string, $allowedHtmlTags );
	}

	/**
	 * Returns the scannable post types.
	 *
	 * @since 1.0.2
	 *
	 * @return array The scannable post types.
	 */
	public function getScannablePostTypes() {
		static $scannablePostTypes = null;
		if ( null !== $scannablePostTypes ) {
			return $scannablePostTypes;
		}

		// We exclude these post types to optimize performance.
		$nonSupportedPostTypes = [ 'attachment' ];
		$scannablePostTypes    = array_diff(
			aioseo()->helpers->getPublicPostTypes( true ),
			$nonSupportedPostTypes
		);

		return $scannablePostTypes;
	}

	/**
	 * Checks whether the current post can be scanned.
	 *
	 * @since 1.0.2
	 *
	 * @param  WP_Post $post The post object.
	 * @return bool          Whether the post is scannable.
	 */
	public function isScannablePost( $post ) {
		if ( ! is_object( $post ) ) {
			return false;
		}

		$postTypes = array_diff( aioseo()->helpers->getPublicPostTypes( true ), [ 'attachment' ] );
		if ( ! in_array( $post->post_type, $postTypes, true ) ) {
			return false;
		}

		if ( ! aioseo()->helpers->isValidPost( $post, aioseo()->helpers->getPublicPostStatuses( true ) ) ) {
			return false;
		}

		return true;
	}
}